import os
import re
import time

import numpy as np

from TDT import *

from IPython.core.debugger import Tracer

def SEV2py(SEV_DIR, *, CHANNEL=0, EVENTNAME='', T1=0, T2=0, FS=0, RANGES=None,
           VERBOSE=0, JUSTNAMES=0):
    '''
    SEV2MAT TDT SEV file format extraction.
    data = SEV2mat(SEV_DIR), where SEV_DIR is a string, retrieves
    all sev data from specified directory in struct format. SEV_DIR can
    also be a single file. SEV files are generated by an RS4 Data Streamer,
    or by enabling the Discrete Files option in the Synapse Stream Data
    Storage gizmo, or by setting the Unique Channel Files option in 
    Stream_Store_MC or Stream_Store_MC2 macro to Yes in OpenEx.

    data    contains all continuous data (sampling rate and raw data)
    
    optional parameters:
        T1          scalar, retrieve data starting at T1 (default = 0 for
                        beginning of recording)
        T2          scalar, retrieve data ending at T2 (default = 0 for end
                        of recording)
        CHANNEL     integer, returns the sev data from specified channel
                        only (default = 0 for all channels)
        RANGES      array of valid time range column vectors
        JUSTNAMES   boolean, retrieve only the valid event names
        EVENTNAME   string, specific event name to retrieve data from
        VERBOSE     boolean, set to false to disable console output
        FS          float, sampling rate override. Useful for lower
                        sampling rates that aren't correctly written into
                        the SEV header.
    '''

    data = structtype()

    MAX_UINT64 = np.iinfo(np.uint64).max
    
    if os.path.isfile(SEV_DIR):
        # treat as single file only
        sev_files = [SEV_DIR]
    elif os.path.isdir(SEV_DIR):
        # treat as directory 
        SEV_DIR = os.path.join(SEV_DIR, '')
        sev_files = get_files(SEV_DIR, '.sev')
    else:
        print('Error: Unable to find sev file or directory:\n\t', SEV_DIR)
        return None
   
    nfiles = len(sev_files)
    if nfiles < 1:
        if JUSTNAMES:
            return []
        print('Warning: no sev files found in', SEV_DIR)
        return None

    if FS > 0:
        print('Using {:.4f} Hz as SEV sampling rate for {}'.format(FS, EVENTNAME))
    
    file_list = []
    for file in sev_files:
        [filename, ext] = os.path.splitext(file)
        [path, name] = os.path.split(filename)
        file_list.append({'fullname':file,
                          'folder':path,
                          'name':name})
    
    
    chan_search = re.compile('_[Cc]h([0-9]*)')
    hour_search = re.compile('-([0-9]*)h')
    name_search = re.compile('_(.{4})_')
    
    # find out what data we think is here
    for file in file_list:
        
        # find channel number
        match_result = chan_search.findall(file['name'])
        if match_result:
            file['chan'] = int(match_result[-1])
        else:
            file['chan'] = -1
        
        # find starting hour
        match_result = hour_search.findall(file['name'])
        if match_result:
            file['hour'] = int(match_result[-1])
        else:
            file['hour'] = 0
        
        # event name of stream
        match_result = name_search.findall(file['name'])
        if match_result:
            file['eventName'] = match_result[-1]
        else:
            file['eventName'] = file['name']
                
        # check file size
        file['data_size'] = os.stat(file['fullname']).st_size - 40
        
        with open(file['fullname'], 'rb') as sev:
            # create and fill streamHeader struct
            streamHeader = structtype()
            
            streamHeader.fileSizeBytes   = np.fromfile(sev, dtype=np.uint64, count=1)[0]
            streamHeader.fileType        = np.fromfile(sev, dtype=np.uint8, count=3)
            streamHeader.fileType        = ''.join([chr(item) for item in streamHeader.fileType])
            streamHeader.fileVersion     = np.fromfile(sev, dtype=np.uint8, count=1)[0]
            streamHeader.eventName       = file['eventName']
            
            if streamHeader.fileVersion < 4:
                
                # prior to v3, OpenEx and RS4 were not setting this properly
                # (one of them was flipping it), so only trust the event name in 
                # header if fileVersion is 3 or higher
                eventName = np.fromfile(sev, dtype=np.uint8, count=4)
                eventName = ''.join([chr(item) for item in eventName])
                if streamHeader.fileVersion >= 3:
                    streamHeader.eventName = eventName
                
                # current channel of stream
                streamHeader.channelNum = np.fromfile(sev, dtype=np.uint16, count=1)[0]
                file['chan'] = streamHeader.channelNum
                # total number of channels in the stream
                streamHeader.totalNumChannels = np.fromfile(sev, dtype=np.uint16, count=1)[0]
                # number of bytes per sample
                streamHeader.sampleWidthBytes = np.fromfile(sev, dtype=np.uint16, count=1)[0]
                reserved = np.fromfile(sev, dtype=np.uint16, count=1)[0]
                
                # data format of stream in lower 3 bits
                dform = np.fromfile(sev, dtype=np.uint8, count=1)[0]
                dform &= 0b111
                streamHeader.dForm = ALLOWED_FORMATS[dform]
                
                # used to compute actual sampling rate
                streamHeader.decimate   = np.fromfile(sev, dtype=np.uint8, count=1)[0]
                streamHeader.rate       = np.fromfile(sev, dtype=np.uint16, count=1)[0]
            else:
                print('Error: unknown version', streamHeader.fileVersion)
                return None
            
            # compute sampling rate
            if streamHeader.fileVersion > 0:
                streamHeader.fs = np.power(2.,(streamHeader.rate - 12)) * 25000000 / streamHeader.decimate
            else:
                # make some assumptions if we don't have a real header
                streamHeader.dForm = 'single'
                streamHeader.fs = 24414.0625
                streamHeader.channelNum = file['chan']
                print('''Warning: {0} has empty header;
assuming {1} ch {2} format {3}
upgrade to OpenEx v2.18 or above\n'''.format(file['name'],
                                             streamHeader.eventName,
                                             streamHeader.channelNum,
                                             streamHeader.dForm))
            
            if FS > 0:
                streamHeader.fs = FS
            
            varname = fix_var_name(streamHeader.eventName)
            file['itemsize'] = np.uint64(np.dtype(streamHeader.dForm).itemsize)
            file['npts'] = file['data_size'] // file['itemsize']
            file['fs'] = streamHeader.fs
            file['dForm'] = streamHeader.dForm
            file['eventName'] = streamHeader.eventName
            file['varName'] = varname

    eventNames = list(set([file['eventName'] for file in file_list]))
    if JUSTNAMES:
        return eventNames

    if T2 > 0:
        validTimeRange = np.array([[T1], [T2]])
    else:
        validTimeRange = np.array([[T1], [np.inf]])

    try:
        len(RANGES)
        validTimeRange = RANGES
    except:
        pass
    
    numRanges = validTimeRange.shape[1]
    if numRanges > 0:
        data['time_ranges'] = validTimeRange
    
    for thisEvent in eventNames:

        if EVENTNAME and EVENTNAME != thisEvent:
            continue
        
        # get files for this event only
        file_list_temp = [file for file in file_list if file['eventName'] == thisEvent]
        
        # extract header info
        fs = file_list_temp[0]['fs']
        eventName = file_list_temp[0]['eventName']
        dForm = file_list_temp[0]['dForm']
        
        chans = [f['chan'] for f in file_list_temp]
        hours = [f['hour'] for f in file_list_temp]
        max_chan = np.max(chans)
        min_chan = np.min(chans)
        max_hour = np.max(hours)
        hour_values = sorted(list(set(hours)))
        
        # preallocate data array
        if CHANNEL > 0:
            try:
                chans.index(CHANNEL)
                search_ch = CHANNEL
            except:
                print('Warning: Channel {0} not found in {1} event'.format(CHANNEL, eventName))
                continue
        else:
            search_ch = min_chan
        
        # determine total samples if there is chunking, and how many samples are in each file
        total_samples = 0
        npts = [np.uint64(0) for i in hour_values]
        for jjj in hour_values:
            ind1 = np.asarray(hours) == jjj
            ind2 = np.asarray(chans) == search_ch
            temp_num = np.where(ind1 & ind2)[0]
            if len(temp_num) < 1:
                print('Error: matching file not found for hour', jjj, 'channel', matching_ch)
                return None
            elif len(temp_num) > 1:
                print('Error: too many matches found for hour', jjj, 'channel', matching_ch)
                return None
            temp_num = temp_num[0]
            npts[jjj] = np.uint64(file_list_temp[temp_num]['npts'])
            total_samples = total_samples + npts[jjj]
        
        # if we are doing time filtering, determine which files we need to read
        # from and how many samples
        absoluteStartSample = np.zeros(numRanges, dtype=np.uint64)
        absoluteEndSample = np.zeros(numRanges, dtype=np.uint64)
        startHourFile = np.zeros(numRanges, dtype=np.uint64)
        endHourFile = np.zeros(numRanges, dtype=np.uint64)
        startHourSamplesToSkip = np.zeros(numRanges, dtype=np.uint64)
        endHourSamplesEnd = np.zeros(numRanges, dtype=np.uint64)
        
        for jj in range(numRanges):
            
            absoluteStartSample[jj] = np.max(np.ceil(validTimeRange[0,jj] * fs), 0) + 1
            if np.isinf(validTimeRange[1,jj]):
                absoluteEndSample[jj] = total_samples
            else:
                absoluteEndSample[jj] = np.minimum(np.max(np.floor(validTimeRange[1,jj] * fs), 0) + 1, total_samples)            
            curr_samples = 0
            for jjj in hour_values:
                if curr_samples <= absoluteStartSample[jj]:
                    startHourSamplesToSkip[jj] = absoluteStartSample[jj] - curr_samples - 1
                    startHourFile[jj] = jjj
                if curr_samples + npts[jjj] >= absoluteEndSample[jj]:
                    endHourSamplesEnd[jj] = absoluteEndSample[jj] - curr_samples
                    endHourFile[jj] = jjj
                    break
                curr_samples = curr_samples + npts[jjj]
        
        # now allocate it
        if CHANNEL > 0:
            channels = [CHANNEL]
        else:
            channels = sorted(list(set(chans)))

        varname = file_list_temp[0]['varName']
        data[varname] = structtype()
        data[varname].data = [[] for i in range(numRanges)]
        for jj in range(numRanges):
            data[varname]['data'][jj] = np.zeros((len(channels), absoluteEndSample[jj] - absoluteStartSample[jj] + np.uint64(1)), dtype=dForm)
        data[varname]['channels'] = channels
        
        # loop through the time ranges
        for ii in range(numRanges):
            
            # loop through the channels
            for chan in channels:
                chanIndex = np.uint64(0)
                
                ind2 = np.asarray(chans) == chan
                
                # loop through the chunks
                for kk in range(startHourFile[ii], endHourFile[ii] + np.uint64(1)):
                    ind1 = np.asarray(hours) == kk
                    file_num = np.where(ind1 & ind2)[0][0]
                    
                    # open file
                    with open(file_list_temp[file_num]['fullname'], 'rb') as f:

                        # skip first 40 bytes from header
                        f.seek(40, os.SEEK_SET)
                        
                        # read rest of file into data array as correct format
                        data[varname]['name'] = eventName
                        data[varname]['fs'] = fs
                        
                        if kk == startHourFile[ii]:
                            firstSample = startHourSamplesToSkip[ii]
                        else:
                            firstSample = 0
                        
                        if kk == endHourFile[ii]:
                            lastSample = endHourSamplesEnd[ii]
                        else:
                            lastSample = MAX_UINT64
                        
                        # skip ahead
                        if firstSample > 0:
                            f.seek(int(firstSample * file_list_temp[file_num]['itemsize']), os.SEEK_CUR)
                        if lastSample == MAX_UINT64:
                            ddd = np.frombuffer(f.read(), dtype=dForm)
                        else:
                            ddd = np.frombuffer(f.read(int((lastSample - firstSample)*file_list_temp[file_num]['itemsize'])), dtype=dForm)
                        readSize = np.uint64(len(ddd))

                        if len(channels) > 1:
                            data[varname].data[ii][chan-1, int(chanIndex):int(chanIndex + readSize)] = ddd
                        else:
                            data[varname].data[ii][0, int(chanIndex):int(chanIndex + readSize)] = ddd
                        chanIndex = chanIndex + readSize
                    
                    if VERBOSE:
                        print(file_list[file_num])
        
        if numRanges == 1:
            data[varname].data = data[varname].data[ii]
        if len(channels) == 1:
            data[varname].data = data[varname].data[ii]
            
    return data
    
if __name__ == '__main__':
    #BLOCK_PATH = 'F:\\OLD_G\\TDT\\Synapse\\Tanks\\IZimptest-170516-094356\\Subject1-170516-094521\\'
    #BLOCK_PATH = 'F:\\OLD_G\\TDT\\Synapse\\Tanks\\IZimptest-170516-094356\\Subject1-170516-094521\\IZimptest-170516-094356_Subject1-170516-094521_EEG2_Ch1.sev'
    #BLOCK_PATH = '\\\\rs4-01001\\data\\Experiment150-181203-154010\\Subject1-181206-122240'
    #BLOCK_PATH = 'C:\\TDT\\Synapse\\Tanks\\Experiment150-181203-154010\\Subject1-181206-122240'
    
    BLOCK_PATH = 'C:\\TDT\\Synapse\\Tanks\\Experiment150-181203-154010\\Subject1-181206-122240'
    #data = SEV2py(BLOCK_PATH, EVENTNAME='EEG1', FS=5000)
    #data = SEV2py(BLOCK_PATH, CHANNEL=5, T2=10)
    #data = SEV2py(BLOCK_PATH, T2=10, CHANNEL=5)
    start = time.time()
    #data = SEV2py(BLOCK_PATH, RANGES=np.array([[5, 3550],[10, 3650]]))    
    #data = SEV2py(BLOCK_PATH, T2=10)
    data = SEV2py(BLOCK_PATH, CHANNEL=5)
    print(time.time() - start, 'elapsed')
    
    #data = SEV2py(BLOCK_PATH, RANGES=np.array([[5, 3550],[10, 3650]]))
    
    #import matplotlib.pyplot as plt
    #plt.plot(data.Ch11.data[0][0,:])
    #plt.show()
    
    #data = SEV2py(BLOCK_PATH, RANGES=np.array([[5, 3550],[10, 3650]]))
    #data = SEV2py(BLOCK_PATH, RANGES=np.array([[5, 3550],[10, 3650]]))
    #data = SEV2py(DEVICE='RS4-01001',TANK='Experiment150-181203-154010', BLOCK='Subject1-181206-122240')


import json
import time
import sys
import re
try:
    import httplib as http # python 2.x
except ImportError:
    import http.client as http # python 3.x

class SynapseAPI:
    Modes = ('Idle', 'Standby', 'Preview', 'Record') #, 'Unknown'

    def __init__(self, server = "localhost", port = 24414):
        self.synCon = http.HTTPConnection(server, port)
        self.lastReqStr = ''
        self.reSueTank = re.compile('subject|user|experiment|tank|block')
        self.demoExperiments = ['demoAudioStim1','demoUser1','demoPCSort','demoBoxSort','demoTetSort','demoChanMap','demoSigSelector','demoSigInjector','demoElecStim','demoFileStim','demoParSeq']
        self.demoRequiredGizmos = {'demoAudioStim1':'aStim1','demoUser1':'TagTest1','demoPCSort':'Neu1','demoBoxSort':'Box1','demoTetSort':'Tet1','demoChanMap':'Map1','demoSigSelector':'Sel1','demoSigInjector':'Inj1','demoElecStim':'eStim1','demoFileStim':'fStim1','demoParSeq':'ParSeq1'}

    def __del__(self):
        self.synCon.close()

    def connect(self):
        self.synCon.close()
        try:
            self.synCon.connect()
        except Exception as e:
            raise Exception('failed to connect to Synapse\n' + str(e))

    def exceptMsg(self):
        retval = ''

        if 'params' in self.lastReqStr:
            retval = '\nSynapse may need to be in non-Idle mode'
        elif self.reSueTank.search(self.lastReqStr) is not None:
            retval = '\nSynapse may need to be in Idle mode'

        return retval

    def getResp(self):
        try:
            resp = self.synCon.getresponse()

            # success
            if resp.status == 200:
                retval = json.loads(resp.read().decode('utf-8'))
            # previous request sent was invalid, why?
            else:
                raise Exception('%s%s' % (resp.reason, self.exceptMsg()))

        except:
            # some HTTP exceptions are such that subsequent communications may fail if we don't re-establish
            self.connect()
            raise Exception('failed to retrieve response from Synapse' + self.exceptMsg())

        return retval

    def sendRequest(self, reqTypeStr, reqStr, reqData = None):
        '''
        reqTypeStr = HTTP methods, e.g. 'GET', 'PUT', 'OPTIONS'
        reqData = JSON formatted data
        '''

        try:
            if reqData is None:
                self.synCon.request(reqTypeStr, reqStr)
            else:
                self.synCon.request(reqTypeStr, reqStr, reqData, {'Content-type' : 'application/json'})

            self.lastReqStr = reqStr

        except:
            self.connect()
            raise Exception('failed to send %s %s to Synapse' % (reqTypeStr, reqStr))

    def sendGet(self, reqStr, respKey = None, reqData = None):
        self.sendRequest('GET', reqStr, reqData)
        resp = self.getResp()

        try:
            if respKey is None:
                retval = resp
            else:
                retval = resp[respKey]

        except:
            retval = None

        return retval

    def sendPut(self, reqStr, reqData):
        self.sendRequest('PUT', reqStr, reqData)
        # we must read and 'clear' response
        # otherwise subsequent HTTP request may fail
        self.getResp()

    def sendOptions(self, reqStr, respKey, reqData = None):
        self.sendRequest('OPTIONS', reqStr, reqData)

        try:
            retval = self.getResp()[respKey]
        except:
            retval = []

        return retval

    def parseJsonString(self, jsonData):
        try:
            retval = str(jsonData)
        except:
            retval = ''

        return retval

    def parseJsonStringList(self, jsonData):
        retval = []
        for value in jsonData:
            retval.append(self.parseJsonString(value))

        return retval

    def parseJsonFloat(self, jsonData, result = []):
        try:
            retval = float(jsonData)
        except:
            retval = 0.0
            # notify caller if interested
            if len(result) > 0:
                result[0] = False

        return retval

    def parseJsonFloatList(self, jsonData, result = []):
        retval = []
        for value in jsonData:
            retval.append(self.parseJsonFloat(value, result))

        return retval

    def parseJsonInt(self, jsonData):
        return int(self.parseJsonFloat(jsonData))

    def getMode(self):
        '''
        -1: Error
         0: Idle
         1: Standby
         2: Preview
         3: Record
        '''

        try:
            retval = self.Modes.index(self.sendGet('/system/mode', 'mode'))
        except:
            retval = -1

        return retval

    def getModeStr(self):
        '''
        '' (Error)
        'Idle'
        'Standby'
        'Preview'
        'Record'
        '''

        retval = self.getMode()
        if retval == -1:
            retval = ''
        else:
            retval = self.Modes[retval]

        return retval

    def setMode(self, mode):
        '''
        mode must be an integer between 0 and 3, inclusive
        '''

        if mode in range(len(self.Modes)):
            self.sendPut('/system/mode', json.dumps({'mode' : self.Modes[mode]}))
        else:
            raise Exception('invalid call to setMode()')

    def setModeStr(self, modeStr):
        '''
        string equivalent of setMode()
        '''

        try:
            mode = self.Modes.index(modeStr)
        except:
            raise Exception('invalid call to setModeStr()')

        self.setMode(mode)

    def issueTrigger(self, id):
        self.sendPut('/trigger/' + str(id), None)

    def getSystemStatus(self):
        retval = {'sysLoad' : 0, 'uiLoad' : 0, 'errorCount' : 0, 'rateMBps' : 0, 'recordSecs' : 0}
        resp = self.sendGet('/system/status')

        sysStat = {'sysLoad' : '', 'uiLoad' : '', 'errors' : '', 'dataRate' : '', 'recDur' : ''}
        for key in resp:
            try:
                sysStat[key] = resp[key]
            except:
                continue

        # Synapse internal keys : user friendly keys
        keyMap = {'sysLoad' : 'sysLoad', 'uiLoad' : 'uiLoad', 'errors' : 'errorCount', 'dataRate' : 'rateMBps', 'recDur' : 'recordSecs'}
        for key in sysStat:
            try:
                if key == 'dataRate':
                    # '0.00 MB/s'
                    retval[keyMap[key]] = float(sysStat[key].split()[0])
                elif key == 'recDur':
                    # 'HH:MM:SSs'
                    recDur = sysStat[key][:-1].split(':')
                    retval[keyMap[key]] = int(recDur[0]) * 3600 + int(recDur[1]) * 60 + int(recDur[2])
                else:
                    retval[keyMap[key]] = int(sysStat[key])

            except:
                continue

        return retval

    def getPersistModes(self):
        return self.parseJsonStringList(self.sendOptions('/system/persist', 'modes'))

    def getPersistMode(self):
        return self.parseJsonString(self.sendGet('/system/persist', 'mode'))

    def setPersistMode(self, modeStr):
        self.sendPut('/system/persist', json.dumps({'mode' : modeStr}))

    def getSamplingRates(self):
        retval = {}
        resp = self.sendGet('/processor/samprate')

        for proc in list(resp.keys()):
            retval[self.parseJsonString(proc)] = self.parseJsonFloat(resp[proc])

        return retval

    def getGizmoParent(self, gizmoName):
        return self.parseJsonString(self.sendGet('/experiment/processor/' + gizmoName, 'processor'))

    def getKnownSubjects(self):
        return self.parseJsonStringList(self.sendOptions('/subject/name', 'subjects'))

    def getKnownUsers(self):
        return self.parseJsonStringList(self.sendOptions('/user/name', 'users'))

    def getKnownExperiments(self):
        return self.parseJsonStringList(self.sendOptions('/experiment/name', 'experiments'))

    def getKnownTanks(self):
        return self.parseJsonStringList(self.sendOptions('/tank/name', 'tanks'))

    def getKnownBlocks(self,):
        return self.parseJsonStringList(self.sendOptions('/block/name', 'blocks'))

    def getCurrentSubject(self):
        return self.parseJsonString(self.sendGet('/subject/name', 'subject'))

    def getCurrentUser(self):
        return self.parseJsonString(self.sendGet('/user/name', 'user'))

    def getCurrentExperiment(self):
        return self.parseJsonString(self.sendGet('/experiment/name', 'experiment'))

    def getCurrentTank(self):
        return self.parseJsonString(self.sendGet('/tank/name', 'tank'))

    def getCurrentBlock(self):
        return self.parseJsonString(self.sendGet('/block/name', 'block'))

    def setCurrentSubject(self, name):
        self.sendPut('/subject/name', json.dumps({'subject' : name}))
        time.sleep(.5)

    def setCurrentUser(self, name, pwd = ''):
        self.sendPut('/user/name', json.dumps({'user' : name, 'pwd' : pwd}))
        time.sleep(.5)

    def setCurrentExperiment(self, name):
        self.sendPut('/experiment/name', json.dumps({'experiment' : name}))
        time.sleep(2)

    def setCurrentTank(self, name):
        self.sendPut('/tank/name', json.dumps({'tank' : name}))
        time.sleep(.5)

    def setCurrentBlock(self, name):
        self.sendPut('/block/name', json.dumps({'block' : name}))
        time.sleep(.5)

    def createTank(self, path):
        self.sendPut('/tank/path', json.dumps({'tank' : path}))
        time.sleep(.5)

    def createSubject(self, name, desc = '', icon = 'mouse'):
        self.sendPut('/subject/name/new', json.dumps({'subject' : name, 'desc' : desc, 'icon' : icon}))
        time.sleep(.5)

    def getGizmoNames(self, synApiOnly = False):
        '''
        if synApiOnly == True:
            return names of objects with any SynapseAPI parameter enabled
        '''
        if synApiOnly:
            return self.parseJsonStringList(self.sendOptions('/gizmos/api', 'gizmos'))
        else:
            return self.parseJsonStringList(self.sendOptions('/gizmos', 'gizmos'))

    def getParameterNames(self, gizmoName):
        return self.parseJsonStringList(self.sendOptions('/params/' + gizmoName, 'parameters'))

    def getParameterInfo(self, gizmoName, paramName):
        info = self.parseJsonStringList(self.sendGet('/params/info/%s.%s' % (gizmoName, paramName), 'info'))
        keys = ('Name', 'Unit', 'Min', 'Max', 'Access', 'Type', 'Array')

        retval = {}
        for i in range(len(keys)):
            key = keys[i]

            try:
                retval[key] = info[i]

                if key == 'Array' and info[i] != 'No' and info[i] != 'Yes':
                    retval[key] = int(info[i])
                elif key == 'Min' or key == 'Max':
                    retval[key] = float(info[i])

            except:
                retval[key] = None

        return retval

    def getParameterSize(self, gizmoName, paramName):
        return self.parseJsonInt(self.sendGet('/params/size/%s.%s' % (gizmoName, paramName), 'value'))

    def getParameterValue(self, gizmoName, paramName):
        value = self.sendGet('/params/%s.%s' % (gizmoName, paramName), 'value')

        didConvert = [True]
        retval = self.parseJsonFloat(value, didConvert)
        
        if not didConvert[0]:
            retval = self.parseJsonString(value)

        return retval

    def getParameterValues(self, gizmoName, paramName, count = -1, offset = 0):
        '''
        if count == -1:
            count = getParameterSize(gizmoName, paramName)
        '''

        count = int(count)
        offset = int(offset)

        if count == -1:
            try:
                count = self.getParameterSize(gizmoName, paramName)
            except:
                count = 1

        values = self.sendGet('/params/%s.%s' % (gizmoName, paramName),
                              'values',
                              json.dumps({'count' : count, 'offset' : offset}))

        # HACK to pass variable by reference
        didConvert = [True]
        retval = self.parseJsonFloatList(values, didConvert)
        
        if not didConvert[0]:
            retval = self.parseJsonStringList(values)
            
        return retval[:min(count, len(retval))]

    def setParameterValue(self, gizmoName, paramName, value):
        self.sendPut('/params/%s.%s' % (gizmoName, paramName), json.dumps({'value' : value}))

    def setParameterValues(self, gizmoName, paramName, values, offset = 0):
        self.sendPut('/params/%s.%s' % (gizmoName, paramName), json.dumps({'offset' : offset, 'values' : values}))

    def getExerimentMemos(self, experiment, memoNum = -1, startNum = -1, endNum = -1, startTime = '-1', endTime = '-1'):
        '''
        memoNum:             return single memo (1 based index)
        startNum / endNum:   filter by log number (nonnegative)
        startTime / endTime: filter by log time stamp (%Y%m%d%H%M%S)

        filtering is inclusive
        startNum / endNum is prioritized over startTime / endTime
        if startNum / startTime > endNum / endTime and endNum / endTime != -1 / '', endNum / endTime is ignored

        all memos for experiment are returned if no filter given
        '''

        reqStr = '/experiment/notes/' + experiment
        if memoNum > -1:
            reqStr += '/%d' % memoNum
        elif startNum > -1 or endNum > -1:
            reqStr += '/range/%d/%d' % (startNum, endNum)
        elif len(startTime) == 14 or len(endTime) == 14:
            reqStr += '/range/%s/%s' % (startTime.replace('-1', '00000000000000'), endTime.replace('-1', '00000000000000'))

        return self.parseJsonStringList(self.sendGet(reqStr, 'notes'))

    def getSubjectMemos(self, subject, memoNum = -1, startNum = -1, endNum = -1, startTime = '-1', endTime = '-1'):
        '''
        memoNum:             return single memo (1 based index)
        startNum / endNum:   filter by log number (nonnegative)
        startTime / endTime: filter by log time stamp (%Y%m%d%H%M%S)

        filtering is inclusive
        startNum / endNum is prioritized over startTime / endTime
        if startNum / startTime > endNum / endTime and endNum / endTime != -1 / '', endNum / endTime is ignored

        all memos for subject are returned if no filter given
        '''

        reqStr = '/subject/notes/' + subject
        if memoNum > -1:
            reqStr += '/%d' % memoNum
        elif startNum > -1 or endNum > -1:
            reqStr += '/range/%d/%d' % (startNum, endNum)
        elif len(startTime) == 14 or len(endTime) == 14:
            reqStr += '/range/%s/%s' % (startTime.replace('-1', '00000000000000'), endTime.replace('-1', '00000000000000'))

        return self.parseJsonStringList(self.sendGet(reqStr, 'notes'))

    def getUserMemos(self, user, memoNum = -1, startNum = -1, endNum = -1, startTime = '-1', endTime = '-1'):
        '''
        memoNum:             return single memo (1 based index)
        startNum / endNum:   filter by log number (nonnegative)
        startTime / endTime: filter by log time stamp (%Y%m%d%H%M%S)

        filtering is inclusive
        startNum / endNum is prioritized over startTime / endTime
        if startNum / startTime > endNum / endTime and endNum / endTime != -1 / '', endNum / endTime is ignored

        all memos for user are returned if no filter given
        '''

        reqStr = '/user/notes/' + user
        if memoNum > -1:
            reqStr += '/%d' % memoNum
        elif startNum > -1 or endNum > -1:
            reqStr += '/range/%d/%d' % (startNum, endNum)
        elif len(startTime) == 14 or len(endTime) == 14:
            reqStr += '/range/%s/%s' % (startTime.replace('-1', '00000000000000'), endTime.replace('-1', '00000000000000'))

        return self.parseJsonStringList(self.sendGet(reqStr, 'notes'))

    def appendExperimentMemo(self, experiment, memo):
        self.sendPut('/experiment/notes', json.dumps({'experiment' : experiment, 'memo' : memo}))

    def appendSubjectMemo(self, subject, memo):
        self.sendPut('/subject/notes', json.dumps({'subject' : subject, 'memo' : memo}))

    def appendUserMemo(self, user, memo):
        self.sendPut('/user/notes', json.dumps({'user' : user, 'memo' : memo}))

    def startDemo(self, name):
        if name not in self.demoExperiments:
            raise Exception('%s is not a valid demo experiment' % name)
        if self.getCurrentExperiment() != name:
            if name not in self.getKnownExperiments():
                raise Exception('Experiment %s not found' % name)
            if self.getModeStr() != 'Idle':
               self.setModeStr('Idle')
            try:
                self.setCurrentExperiment(name)
            except:
                raise Exception('Experiment %s not selected' % name)

        if self.demoRequiredGizmos[name] not in self.getGizmoNames():
            raise Exception('Required gizmo %s not found' % self.demoRequiredGizmos[name])

        if self.getModeStr() == 'Idle':
            self.setPersistMode('Fresh')
            self.setModeStr('Record')



__version__ = '0.1.6'

import os
import re

import numpy as np

# Tank event types (tsqEventHeader.type)
EVTYPE_UNKNOWN  = int('00000000', 16)
EVTYPE_STRON    = int('00000101', 16)
EVTYPE_STROFF	= int('00000102', 16)
EVTYPE_SCALAR	= int('00000201', 16)
EVTYPE_STREAM	= int('00008101', 16)
EVTYPE_SNIP		= int('00008201', 16)
EVTYPE_MARK		= int('00008801', 16)
EVTYPE_HASDATA	= int('00008000', 16)
EVTYPE_UCF		= int('00000010', 16)
EVTYPE_PHANTOM	= int('00000020', 16)
EVTYPE_MASK		= int('0000FF0F', 16)
EVTYPE_INVALID_MASK	= int('FFFF0000', 16)

EVMARK_STARTBLOCK	= int('0001', 16)
EVMARK_STOPBLOCK	= int('0002', 16)

DFORM_FLOAT		 = 0
DFORM_LONG		 = 1
DFORM_SHORT		 = 2
DFORM_BYTE		 = 3
DFORM_DOUBLE	 = 4
DFORM_QWORD		 = 5
DFORM_TYPE_COUNT = 6

ALLOWED_FORMATS = [np.float32, np.int32, np.int16, np.int8, np.float64, np.int64]
ALLOWED_EVTYPES = ['all','epocs','snips','streams','scalars']

MASKS = {1:0x01, 2:0x03, 4:0x0f, 8:0xff}

def get_files(dir, ext):
    result = []
    for file in os.listdir(dir):
        if file.endswith(ext):
            result.append(os.path.join(dir, file))
    return result

def fix_var_name(var_str, verbose=False):
    if not var_str[0].isalnum():
        var_str = 'x' + var_str
    fixed_name = re.sub('\W|^(?=\d)','_', var_str)
    if fixed_name != var_str and verbose:
        print('info: {0} is not a valid Python variable name, changing to {1}'.format(var_str, fixed_name))
    return fixed_name

class StructType(dict):

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)
        
    def __repr__(self):
        if len(self.__dict__.items()) < 1:
            return '{}'

        parts = []
        for k in self.__dict__.keys():
            if isinstance(self.__dict__[k], StructType):
                rrr = repr(self.__dict__[k])
                parts.append(k + "\t[struct]")
            else:
                parts.append(k + ":\t" + repr(self.__dict__[k]))
        result = '\n'.join(parts)
        return result
    
    def __bool__(self):
        return len(self.__dict__.keys()) > 0
    
    def __getitem__(self, key):
        val = getattr(self, key)
        return val
    
    def __setitem__(self, key, val):
        return setattr(self, key, val)
    
    def keys(self):
        return self.__dict__.keys()
    
    def items(self):
        return self.__dict__.items()
    
    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

# converts incoming packet into sort code counts
def unpack(data, sort_codes=2, bits_per_bin=4):
    mask = MASKS[bits_per_bin]
    return [np.bitwise_and(data >> bits_per_bin * x, mask) for x in range(sort_codes)]

def download_demo_data():
    import os
    if not os.path.exists('data.zip'):
        import urllib.request
        import sys
        import time

        url = 'https://www.tdt.com/support/examples/TDTExampleData.zip'
        print('downloading demo data...')

        def reporthook(count, block_size, total_size):
            global start_time
            if count == 0:
                start_time = time.time()
                return
            duration = time.time() - start_time
            progress_size = int(count * block_size)
            if duration > 0:
                speed = int(progress_size / (1024 * duration))
                percent = min(int(count * block_size * 100 / total_size), 100)
                sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds elapsed" %
                                (percent, progress_size / (1024 * 1024), speed, duration))
                sys.stdout.flush()

        urllib.request.urlretrieve(url, 'data.zip', reporthook)
        print()

    if not os.path.exists('data'):
        try:
            print('unzipping demo data...')
            import zipfile
            zip_ref = zipfile.ZipFile('data.zip', 'r')
            zip_ref.extractall('data')
            zip_ref.close()
        except:
            print('problem with zip, downloading again')
            os.remove('data.zip')
            return download_demo_data()

    print('demo data ready')

from .TDTbin2py import read_block, read_sev
from .TDTfilter import epoc_filter
from .SynapseAPI import SynapseAPI
from .TDTUDP import TDTUDP


import socket
import struct
import time

import tdt

class TDTUDP():
    def __init__(self, host='10.1.0.100', send_type=float, recv_type=float, verbose=False):

        self.NPACKETS = -1
        self.send_header = None

        self.TDT_UDP_HOSTNAME = host
        self.send_type = send_type
        self.recv_type = recv_type
        self.verbose = verbose

        # UDP command constants
        self.CMD_SEND_DATA        = 0x00
        self.CMD_GET_VERSION      = 0x01
        self.CMD_SET_REMOTE_IP    = 0x02
        self.CMD_FORGET_REMOTE_IP = 0x03

        # Important: the RZ UDP interface port is fixed at 22022
        self.UDP_PORT = 22022

        # create a UDP socket object
        self.sock = socket.socket( socket.AF_INET, socket.SOCK_DGRAM ) # UDP

        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
        # connect the PC to the target UDP interface
        self.sock.connect((self.TDT_UDP_HOSTNAME, self.UDP_PORT))

        # configure the header. Notice that it includes the header
        # information followed by the command 2 (set remote IP)
        # and 0 (no data packets for header).
        packet = struct.pack('4B', 0x55, 0xAA, self.CMD_SET_REMOTE_IP, 0)

        # Sends the packet to the UDP interface, setting the remote IP
        # address of the UDP interface to the host PC
        self.sock.send(packet)

    def recv(self):

        # receive a data packet from the UDP interface
        packet = self.sock.recv(1024)
        if len(packet) < 2:
            return None

        # check that magic number is in first position of packet
        if struct.unpack('>h', packet[:2]) != (0x55AA,):
            print('bad header')
            return None

        # unpack it
        npack = (len(packet)-2) / 4
        if self.recv_type == int:
            fmt = '>%dL' % npack
        elif self.recv_type == float:
            fmt = '>%df' % npack
        result = struct.unpack(fmt, packet[4:])
        
        if self.verbose:
            print('received packet', result)
        
        return result

    def send(self, data=[]):

        if len(data) < 1:
            return None

        if len(data) != self.NPACKETS:
            self.NPACKETS = len(data)

            # configure the header
            self.send_header = struct.pack('4B', 0x55, 0xAA, self.CMD_SEND_DATA, self.NPACKETS)

        # append the 32-bit words to the header
        # '>' in the format string is used to force big-endian
        if self.send_type == int:
            packet = struct.pack(">%di" % self.NPACKETS, *(i for i in data))
        elif self.send_type == float:
            packet = struct.pack(">%df" % self.NPACKETS, *(i for i in data))

        # send the data packet to the UDP interface.
        if self.verbose:
            print('sending packet', data, '...')
        self.sock.send(self.send_header + packet)

if __name__ == '__main__':
    import time

    #host = 'localhost'
    host = '10.10.10.167'
    
    udp = TDTUDP(host=host, send_type=int, recv_type=int)

    # SEND ONLY EXAMPLE
    if 0:
        SEND_PACKETS = 1
        ct = 0
        while 1:
            ct += 1
            fakedata = range(ct % 10, SEND_PACKETS + ct % 10)
            if udp.send_type == float:
                fakedata = [x * 2. for x in fakedata]
            udp.send(fakedata)
            time.sleep(.1) # slow it down a bit

    # RECEIVE ONLY EXAMPLE
    if 0:
        while 1:
            data = udp.recv()
            
            '''
            print('RAW DATA:', end='\t')
            for d in data:
                print(d, end='\t')
            print(8*'\t', end='\r')
            '''
            print(data)
            continue
            # if looking at binner packets, extract sort codes
            unpacked_data = tdt.unpack(data, sort_codes=2, bits_per_bin=4)
            print('UNPACKED DATA:', unpacked_data, end='\t\t\t\r')
            
            # extract a single channel/sort from the unpacked data
            channel = 16
            sort_code = 1
            #print('CHANNEL:', channel, 'SORT:', sort_code, '\t', unpacked_data[sort_code-1][channel-1], end='\t\t\t\r')

    # SEND AND RECEIVE EXAMPLE
    if 1:
        SEND_PACKETS = 8
        ct = 0
        while 1:
            ct += 1
            fakedata = range(ct % 10, SEND_PACKETS + ct % 10)
            if udp.send_type == float:
                fakedata = [x * 2. for x in fakedata]
            
            data = udp.recv()
            print(data)

            udp.send(fakedata)